# coding: utf-8
from enum import Enum
from enum import IntEnum
from enum import unique


# !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
#
#                                  ВНИМАНИЕ!
#
# При изменении указанных здесь констант следует актуализировать их в
# ObjectCoordination.js.
#
# !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

# -----------------------------------------------------------------------------


@unique
class MoveDirection(IntEnum):

    """Направления перемещения этапов в маршруте согласования."""

    #: Вверх.
    UP = 0

    #: Вниз.
    DOWN = 1
# -----------------------------------------------------------------------------


@unique
class ReviewResult(IntEnum):

    """Результаты рассмотрения объекта согласования."""

    #: Согласовано.
    AGREED = 0

    #: На доработку.
    #:
    #: Данный результат рассмотрения позволяет возобновить маршрут
    #: согласования (см. :func:`~m3_object_coordination.utils.resume_route`).
    NEED_CHANGES = 1

    #: Отклонено.
    #:
    #: Данный результат рассмотрения делает невозможным возобновление
    #: согласования (см. :func:`~m3_object_coordination.utils.resume_route`).
    REJECTED = 2


#: Текстовые наименования результатов рассмотрения объектов согласования.
REVIEW_RESULT_NAMES = {
    ReviewResult.AGREED: 'Согласовано',
    ReviewResult.NEED_CHANGES: 'На доработку',
    ReviewResult.REJECTED: 'Отклонено',
}
# -----------------------------------------------------------------------------


@unique
class RouteState(IntEnum):

    """Состояния маршрута."""

    #: Подготовка.
    #:
    #: В этом состоянии находятся маршруты, у которых еще нет этапов, либо
    #: содержащие этапы без согласующих.
    PREPARATION = 1

    #: В ожидании.
    #:
    #: Данное состояние соответствует маршрутам с созданными этапами и
    #: согласующими в каждом их них, при этом ни в одном из этапов маршрута не
    #: заполнена плановая дата исполнения (поле ``planning_date``).
    WAITING = 2

    #: Исполняется.
    #:
    #: Данное состояние соответствует маршрутам с созданными этапами и
    #: согласующими в каждом их них, при этом хотя бы в одном из этапов
    #: маршрута должна быть заполнена плановая дата исполнения (поле
    #: ``planning_date``) и хотя бы у одного этапа не заполнена фактическая
    #: дата исполнения.
    EXECUTING = 3

    #: Согласован.
    #:
    #: Состояние соответствует маршруту, у которого все этапы согласованы.
    APPROOVED = 4

    #: Отправлен на доработку.
    #:
    #: Состояние соответствует маршруту, на одном из этапов которого есть
    #: результат рассмотрения "На доработку" и нет результатов "Отклонено".
    NEED_CHANGES = 5

    #: Отклонен.
    #:
    #: Состояние соответствует маршруту, в котором есть результат рассмотрения
    #: "Отклонено".
    REJECTED = 6


#: Текстовые наименования состояний маршрута.
ROUTE_STATE_NAMES = {
    RouteState.PREPARATION: 'Подготовка',
    RouteState.WAITING: 'В ожидании',
    RouteState.EXECUTING: 'Исполняется',
    RouteState.APPROOVED: 'Согласован',
    RouteState.NEED_CHANGES: 'Отправлен на доработку',
    RouteState.REJECTED: 'Отклонен',
}


#: Статусы маршрута, указывающие на завершение процесса согласования.
FINAL_ROUTE_STATES = (
    RouteState.APPROOVED,
    RouteState.NEED_CHANGES,
    RouteState.REJECTED,
)


#: Статусы маршрута, указывающие на отрицательный результат согласования.
NEGATIVE_ROUTE_STATES = (
    RouteState.NEED_CHANGES,
    RouteState.REJECTED,
)


#: Статусы маршрута, позволяющие выполнить его сброс.
#:
#: .. seealso::
#:
#:    :func:`~m3_object_coordination.utils.reset_route`
RESETABLE_ROUTE_STATES = (
    RouteState.EXECUTING,
    RouteState.APPROOVED,
    RouteState.NEED_CHANGES,
    RouteState.REJECTED,
)
# -----------------------------------------------------------------------------


@unique
class PhaseState(IntEnum):

    """Состояния этапов согласования."""

    #: В ожидании.
    #:
    #: Состояние соответствует этапу, согласование которого еще не начато, при
    #: этом предыдущие этапы не были отклонены или отправлены на доработку.
    WAITING = 1

    #: Исполняется.
    #:
    #: Состояние соответствует этапу, который находится в процессе рассмотрения
    #: согласующими.
    EXECUTING = 2

    #: Исполненный.
    #:
    #: Состояние соответствует этапу, в котором все согласующие рассмотрели
    #: предмет согласования с результатом "Согласовано".
    EXECUTED = 3

    #: Остановлен.
    #:
    #: Состояние соответствует этапу, который был отклонен, либо отправлен на
    #: доработку.
    STOPPED = 4

    #: Заблокированный.
    #:
    #: Состояние соответствует этапу, согласование которого не было начато, при
    #: этом один из предыдущих этапов был отклонен, либо отправлен на
    #: доработку.
    BLOCKED = 5


#: Текстовые наименования состояний этапов согласования.
PHASE_STATE_NAMES = {
    PhaseState.WAITING: 'В ожидании',
    PhaseState.EXECUTING: 'Исполняемый',
    PhaseState.EXECUTED: 'Исполненный',
    PhaseState.STOPPED: 'Остановлен',
    PhaseState.BLOCKED: 'Заблокированный',
}
# -----------------------------------------------------------------------------


@unique
class ContextParser(Enum):

    """Имена внутренних парсеров параметров контекста."""

    #: Кортеж из целых чисел.
    INT_TUPLE = 'm3-object-coordination:int-tuple'

    #: Целое число, либо None.
    INT_OR_NONE = 'm3-object-coordination:int-or-none'
# -----------------------------------------------------------------------------
